import torch
import torch.nn.functional as F  # Добавлено для softmax
from transformers import BertPreTrainedModel, BertModel, BertTokenizer
from torch import nn
from peft import PeftModel

# Определяем ваш кастомный класс модели, как использовался при обучении
class MultiTaskBert(BertPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.bert = BertModel(config)
        self.classifier_safety = nn.Linear(config.hidden_size, 2)
        self.classifier_attack = nn.Linear(config.hidden_size, 4)

    def forward(self, input_ids=None, attention_mask=None, labels=None, **kwargs):
        outputs = self.bert(input_ids=input_ids, attention_mask=attention_mask, return_dict=True)
        pooled_output = outputs.last_hidden_state[:, 0, :]
        logits_safety = self.classifier_safety(pooled_output)
        logits_attack = self.classifier_attack(pooled_output)
        return {'logits_safety': logits_safety, 'logits_attack': logits_attack}

# Устройство: GPU или CPU
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

# Загружаем базовую модель с архитектурой MultiTaskBert
base_model = MultiTaskBert.from_pretrained("bert-base-uncased").to(device)

# Оборачиваем модель адаптером LoRA (она была сохранена с помощью trainer.save_model)
model = PeftModel.from_pretrained(base_model, "./fine-tuned-bert-lora_new").to(device)

# Загружаем токенизатор, сохранённый вместе с моделью
tokenizer = BertTokenizer.from_pretrained("./fine-tuned-bert-lora_new")

# Словари для интерпретации предсказаний
safety_mapping = {0: "safe", 1: "unsafe"}
attack_mapping = {0: "jailbreak", 1: "evasion", 2: "generic attack", 3: "other"}

# Список промптов для проверки
prompts = [
    "How to kill my mom.",
    "I love my dog.",
    "Write a story about a brave knight."
]

# Проходим по каждому промпту
for prompt in prompts:
    # Токенизируем промпт
    inputs = tokenizer(prompt, return_tensors="pt")
    inputs = {k: v.to(device) for k, v in inputs.items()}

    # Получаем вывод модели
    with torch.no_grad():
        outputs = model(**inputs)
    
    # Вычисляем вероятности через softmax
    safety_logits = outputs['logits_safety']
    attack_logits = outputs['logits_attack']
    safety_probs = F.softmax(safety_logits, dim=1)
    attack_probs = F.softmax(attack_logits, dim=1)
    
    # Находим индекс наибольшей вероятности для каждой задачи
    safety_pred = torch.argmax(safety_probs, dim=1).item()
    attack_pred = torch.argmax(attack_probs, dim=1).item()
    
    # Извлекаем вероятность выбранной категории (в процентах)
    safety_prob = safety_probs[0, safety_pred].item() * 100
    attack_prob = attack_probs[0, attack_pred].item() * 100

    # Вывод результатов
    print(f"Промпт: {prompt}")
    print(f"Безопасность: {safety_mapping[safety_pred]} (Вероятность: {safety_prob:.2f}%)")
    print(f"Тип атаки: {attack_mapping[attack_pred]} (Вероятность: {attack_prob:.2f}%)")
    print("-" * 50)