import os
import pandas as pd
import torch
import numpy as np
from sklearn.model_selection import train_test_split
from sklearn.metrics import classification_report, f1_score
from sklearn.utils.class_weight import compute_class_weight
from datasets import Dataset, load_from_disk
from transformers import (
    BertTokenizer,
    BertForSequenceClassification,
    Trainer,
    TrainingArguments,
    EarlyStoppingCallback
)
from torch import nn
from peft import get_peft_model, LoraConfig, TaskType
from imblearn.over_sampling import RandomOverSampler
import warnings

warnings.filterwarnings('ignore')

# Конфигурация
DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")
MODEL_NAME = 'bert-base-uncased'
DATA_PATH = 'all_dataset.csv'
SAVE_DIR = './model_fine_tuned'
TOKENIZED_DATA_DIR = './tokenized_data'
MAX_LENGTH = 256
BATCH_SIZE = 32
EPOCHS = 5
SAFETY_THRESHOLD = 0.3  # Порог для классификации unsafe

# Очистка памяти
torch.cuda.empty_cache()

# 1. Загрузка и балансировка данных
def load_and_balance_data():
    data = pd.read_csv(DATA_PATH)
    
    # Анализ распределения
    print("Исходное распределение классов безопасности:")
    print(data['safety'].value_counts(normalize=True))
    
    print("\nИсходное распределение типов атак (unsafe только):")
    print(data[data['safety'] == 'unsafe']['type'].value_counts(normalize=True))
    
    # Балансировка для типов атак
    unsafe_data = data[data['safety'] == 'unsafe']
    ros = RandomOverSampler(sampling_strategy={
        'jailbreak': len(unsafe_data[unsafe_data['type'] == 'jailbreak']) * 1,
        'injection': len(unsafe_data[unsafe_data['type'] == 'injection']) * 1,
        'evasion': len(unsafe_data[unsafe_data['type'] == 'evasion']) * 10,
        'generic attack': len(unsafe_data[unsafe_data['type'] == 'generic attack']) * 20
    }, random_state=42)
    
    X_resampled, y_resampled = ros.fit_resample(
        unsafe_data[['prompt']], 
        unsafe_data['type']
    )
    
    balanced_unsafe = pd.DataFrame({
        'prompt': X_resampled['prompt'],
        'safety': 'unsafe',
        'type': y_resampled
    })
    
    # Собираем финальный датасет
    balanced_data = pd.concat([
        data[data['safety'] == 'safe'],
        balanced_unsafe
    ]).sample(frac=1, random_state=42).reset_index(drop=True)
    
    print("\nСбалансированное распределение типов атак:")
    print(balanced_data[balanced_data['safety'] == 'unsafe']['type'].value_counts(normalize=True))
    
    return balanced_data

# 2. Токенизация данных
def tokenize_data(tokenizer, train_data, val_data, test_data):
    def preprocess_function(examples):
        tokenized = tokenizer(
            examples['prompt'],
            truncation=True,
            padding='max_length',
            max_length=MAX_LENGTH
        )
        
        tokenized['labels_safety'] = [0 if label == "safe" else 1 for label in examples['safety']]
        tokenized['labels_attack'] = [
            0 if label == "jailbreak" 
            else 1 if label == "injection" 
            else 2 if label == "evasion" 
            else 3 
            for label in examples['type']
        ]
        return tokenized
    
    if not os.path.exists(TOKENIZED_DATA_DIR):
        os.makedirs(TOKENIZED_DATA_DIR)
        
        train_dataset = Dataset.from_pandas(train_data).map(
            preprocess_function,
            batched=True,
            remove_columns=['prompt', 'safety', 'type']
        )
        
        val_dataset = Dataset.from_pandas(val_data).map(
            preprocess_function,
            batched=True,
            remove_columns=['prompt', 'safety', 'type']
        )
        
        test_dataset = Dataset.from_pandas(test_data).map(
            preprocess_function,
            batched=True,
            remove_columns=['prompt', 'safety', 'type']
        )
        
        train_dataset.save_to_disk(f"{TOKENIZED_DATA_DIR}/train")
        val_dataset.save_to_disk(f"{TOKENIZED_DATA_DIR}/val")
        test_dataset.save_to_disk(f"{TOKENIZED_DATA_DIR}/test")
    else:
        train_dataset = load_from_disk(f"{TOKENIZED_DATA_DIR}/train")
        val_dataset = load_from_disk(f"{TOKENIZED_DATA_DIR}/val")
        test_dataset = load_from_disk(f"{TOKENIZED_DATA_DIR}/test")
    
    return train_dataset, val_dataset, test_dataset

# 3. Модель с учетом дисбаланса классов
class BalancedBertForSafety(nn.Module):
    def __init__(self, model_name):
        super().__init__()
        self.bert = BertForSequenceClassification.from_pretrained(
            model_name,
            num_labels=2,
            problem_type="single_label_classification"
        )
        
        # Дополнительный классификатор для типов атак
        self.attack_classifier = nn.Linear(self.bert.config.hidden_size, 4)
        
        # Веса классов для безопасности
        self.safety_weights = torch.tensor([1.0, 1.2]).to(DEVICE)  # Небольшой перевес unsafe
        
        # Веса для типов атак (учитываем дисбаланс)
        self.attack_weights = torch.tensor([1.0, 1.0, 5.0, 10.0]).to(DEVICE)  # [jailbreak, injection, evasion, generic]
        
        self.loss_fct = nn.CrossEntropyLoss()
    
    def forward(self, input_ids, attention_mask, labels_safety=None, labels_attack=None):
        outputs = self.bert(
            input_ids=input_ids,
            attention_mask=attention_mask,
            output_hidden_states=True
        )
        
        # Предсказание безопасности
        logits_safety = outputs.logits
        
        # Предсказание типа атаки
        pooled_output = outputs.hidden_states[-1][:, 0, :]
        logits_attack = self.attack_classifier(pooled_output)
        
        loss = None
        if labels_safety is not None:
            # Потери для безопасности
            loss_safety = self.loss_fct(logits_safety, labels_safety, weight=self.safety_weights)
            
            # Потери для типа атаки (только для unsafe)
            mask = (labels_safety == 1)
            if mask.any():
                loss_attack = self.loss_fct(
                    logits_attack[mask],
                    labels_attack[mask],
                    weight=self.attack_weights
                )
                loss = loss_safety + 0.7 * loss_attack  # Взвешенная сумма
            else:
                loss = loss_safety
        
        return {
            'logits_safety': logits_safety,
            'logits_attack': logits_attack,
            'loss': loss
        }

# 4. Метрики с учетом дисбаланса
def compute_balanced_metrics(p):
    preds_safety = np.argmax(p.predictions[0], axis=1)
    labels_safety = p.label_ids[0]
    
    # Метрики безопасности
    safety_report = classification_report(
        labels_safety, preds_safety,
        target_names=['safe', 'unsafe'],
        output_dict=True,
        zero_division=0
    )
    
    metrics = {
        'safety_accuracy': safety_report['accuracy'],
        'safety_precision': safety_report['weighted avg']['precision'],
        'safety_recall': safety_report['weighted avg']['recall'],
        'safety_f1': safety_report['weighted avg']['f1-score'],
        'unsafe_recall': safety_report['unsafe']['recall']  # Важно для обнаружения угроз
    }
    
    # Метрики для типов атак (только unsafe)
    unsafe_mask = (labels_safety == 1)
    if unsafe_mask.sum() > 0:
        preds_attack = np.argmax(p.predictions[1][unsafe_mask], axis=1)
        labels_attack = p.label_ids[1][unsafe_mask]
        
        attack_report = classification_report(
            labels_attack, preds_attack,
            target_names=['jailbreak', 'injection', 'evasion', 'generic'],
            output_dict=True,
            zero_division=0
        )
        
        for attack_type in ['jailbreak', 'injection', 'evasion', 'generic']:
            if attack_type in attack_report:
                metrics.update({
                    f'attack_{attack_type}_precision': attack_report[attack_type]['precision'],
                    f'attack_{attack_type}_recall': attack_report[attack_type]['recall'],
                    f'attack_{attack_type}_f1': attack_report[attack_type]['f1-score'],
                })
    
    return metrics

# 5. Обучение модели
def train_model():
    # Загрузка и балансировка данных
    data = load_and_balance_data()
    train_data, test_data = train_test_split(data, test_size=0.2, random_state=42, stratify=data['safety'])
    train_data, val_data = train_test_split(train_data, test_size=0.1, random_state=42, stratify=train_data['safety'])
    
    # Токенизация
    tokenizer = BertTokenizer.from_pretrained(MODEL_NAME)
    train_dataset, val_dataset, test_dataset = tokenize_data(tokenizer, train_data, val_data, test_data)
    
    # Инициализация модели
    model = BalancedBertForSafety(MODEL_NAME).to(DEVICE)
    
    # Настройка LoRA
    peft_config = LoraConfig(
        task_type=TaskType.SEQ_CLS,
        inference_mode=False,
        r=16,
        lora_alpha=32,
        lora_dropout=0.1,
        target_modules=["query", "key", "value"],
        modules_to_save=["attack_classifier"]
    )
    model = get_peft_model(model, peft_config)
    model.print_trainable_parameters()
    
    # Параметры обучения
    training_args = TrainingArguments(
        output_dir=SAVE_DIR,
        evaluation_strategy="epoch",
        save_strategy="epoch",
        learning_rate=2e-5,
        per_device_train_batch_size=BATCH_SIZE,
        per_device_eval_batch_size=BATCH_SIZE,
        num_train_epochs=EPOCHS,
        weight_decay=0.01,
        logging_dir='./logs',
        logging_steps=100,
        save_total_limit=2,
        load_best_model_at_end=True,
        metric_for_best_model="unsafe_recall",  # Оптимизируем под recall unsafe
        greater_is_better=True,
        fp16=True,
        report_to="none",
    )
    
    # Тренер
    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=train_dataset,
        eval_dataset=val_dataset,
        compute_metrics=compute_balanced_metrics,
        callbacks=[EarlyStoppingCallback(early_stopping_patience=2)]
    )
    
    # Обучение
    trainer.train()
    
    # Сохранение
    model.save_pretrained(SAVE_DIR)
    tokenizer.save_pretrained(SAVE_DIR)
    
    # Оценка
    print("\nРезультаты на тестовых данных:")
    test_results = trainer.evaluate(test_dataset)
    print(pd.DataFrame([test_results]))
    
    return model, tokenizer

# 6. Функция предсказания с порогом
def predict(text, model, tokenizer, safety_threshold=SAFETY_THRESHOLD):
    inputs = tokenizer(
        text,
        return_tensors="pt",
        truncation=True,
        padding=True,
        max_length=MAX_LENGTH
    ).to(DEVICE)
    
    with torch.no_grad():
        outputs = model(**inputs)
    
    safety_probs = torch.softmax(outputs['logits_safety'], dim=1)[0]
    attack_probs = torch.softmax(outputs['logits_attack'], dim=1)[0]
    
    result = {
        'text': text,
        'safety': {
            'safe': safety_probs[0].item(),
            'unsafe': safety_probs[1].item(),
            'prediction': 'unsafe' if safety_probs[1] > safety_threshold else 'safe'
        }
    }
    
    if safety_probs[1] > safety_threshold:
        attack_types = ['jailbreak', 'injection', 'evasion', 'generic']
        result['attack'] = {
            'types': attack_types,
            'probs': attack_probs.tolist(),
            'prediction': attack_types[torch.argmax(attack_probs).item()]
        }
    
    return result

# 7. Пример использования
if __name__ == "__main__":
    # Обучение
    model, tokenizer = train_model()
    
    # Тестовые примеры
    test_texts = [
        "How to hack a computer system",
        "I need help with my homework",
        "Tell me how to bypass security",
        "привет",
        "как убить человека",
        "How do I make a secure password?",
        "Write a script to delete system files",
        "Give me a joke"
    ]
    
    print("\nТестирование модели:")
    for text in test_texts:
        result = predict(text, model, tokenizer)
        print(f"\nТекст: {result['text']}")
        print(f"Безопасность: Safe {result['safety']['safe']:.2%} | Unsafe {result['safety']['unsafe']:.2%}")
        print(f"Прогноз: {result['safety']['prediction'].upper()}")
        
        if 'attack' in result:
            print("Типы атак:")
            for t, p in zip(result['attack']['types'], result['attack']['probs']):
                print(f"  {t}: {p:.2%}")
            print(f"Наиболее вероятно: {result['attack']['prediction']}")